@@ -1,6 +1,6 @@
-import heapq
-
-
+# import heapq
+# 
+# 
 # class PriorityQueue:
 #     def __init__(self, data: list = None, min_heap: bool = True):
 #         self.min_heap = min_heap
@@ -35,13 +35,15 @@ import heapq
 #         if len(self) <= 0:
 #             raise IndexError("Pop on an empty queue")
 #         return heapq.heappop(self.heap) if self.min_heap else heapq._heappop_max(self.heap)
+import time
+import threading
 
 
 class PriorityQueue:
     def __init__(self, init_queue: list = None, min_heap: bool = True):
         self.min_heap = min_heap
         self.heap = list()  # (priority, key)
-        if init_queue:
+        if init_queue is not None:
             for item in init_queue:
                 self.push(item)
 
@@ -51,12 +53,6 @@ class PriorityQueue:
     def __str__(self):
         return str(self.heap)
 
-    def peek(self):
-        """ O(1) """
-        if len(self) <= 0:
-            raise IndexError("Peek at an empty queue")
-        return self.heap[0][0] if len(self.heap[0]) == 1 else self.heap[0]  # priority / priority, key
-
     def push(self, *priority_n_key):
         """ O(logN) """
         if len(priority_n_key) == 0:
@@ -67,6 +63,23 @@ class PriorityQueue:
             self.heap.append((priority_n_key[0], priority_n_key[1]))
         self._sift_down(len(self.heap) - 1)
 
+    def peek(self):
+        """ O(1) """
+        if len(self) <= 0:
+            raise IndexError("Peek at an empty queue")
+        return self.heap[0][0] if len(self.heap[0]) == 1 else self.heap[0]  # priority / priority, key
+
+    def top_n(self, n: int = None) -> list():
+        copy = PriorityQueue(min_heap=self.min_heap)
+        copy.heap = self.heap.copy()
+        top_n_list, count = list(), 0
+        while len(copy) > 0:
+            if n is not None and count >= n:
+                break
+            top_n_list.append(copy.pop())
+            count += 1
+        return top_n_list
+
     def pop(self):
         """ O(logN) """
         if len(self) <= 0:
@@ -118,7 +131,7 @@ class PriorityMap(PriorityQueue):
     def __init__(self, init_map: dict = None, min_heap: bool = True):
         super().__init__(min_heap=min_heap)
         self.map = dict() # <key, heap_index>
-        if init_map:
+        if init_map is not None:
             for key, priority in init_map.items():
                 self.push(priority, key)
 
@@ -137,6 +150,18 @@ class PriorityMap(PriorityQueue):
             raise IndexError("Peek at an empty queue")
         return self.heap[0]
 
+    def top_n(self, n: int = None) -> list():
+        copy = PriorityMap(min_heap=self.min_heap)
+        copy.heap = self.heap.copy()
+        copy.map = self.map.copy()
+        top_n_list, count = list(), 0
+        while len(copy) > 0:
+            if n is not None and count >= n:
+                break
+            top_n_list.append(copy.pop())
+            count += 1
+        return top_n_list
+
     def push(self, *priority_n_key):
         """ O(logN) """
         priority, key = priority_n_key[0] if type(priority_n_key[0]) is tuple else (priority_n_key[0], priority_n_key[1])
@@ -228,5 +253,71 @@ class PriorityMap(PriorityQueue):
         self.map[new_item[1]] = index
 
 
+class BlockingPriorityQueue:
+    def __init__(self, max_size: int = None, min_heap: bool = True):
+        self.priority_queue = PriorityQueue(min_heap=min_heap)
+        self.max_size = max_size
+        self.lock = threading.Lock()
+        self.write_condition = threading.Condition()
+        self.read_condition = threading.Condition()
+        if max_size is not None and max_size < 0:
+            raise ValueError(f"Negative queue size found: {max_size}")
+
+    def __len__(self):
+        self.lock.acquire()
+        size = len(self.priority_queue)
+        self.lock.release()
+        return size
+
+    @staticmethod
+    def _notify(condition):
+        condition.acquire()
+        condition.notify()
+        condition.release()
+
+    @staticmethod
+    def _wait(condition, timeout=None):
+        condition.acquire()
+        condition.wait(timeout=timeout)
+        condition.release()
+
+    def push(self, item, block: bool = True, timeout: int = None):
+        if self.max_size is not None and len(self) > self.max_size:
+            if block:  # Block and wait till not full
+                BlockingPriorityQueue._wait(self.write_condition, timeout)
+        self.lock.acquire()
+        try:
+            if self.max_size is not None and len(self.priority_queue) > self.max_size:
+                # do not use len(self) since it has already been locked
+                raise IndexError(f"Queue is full")
+            self.priority_queue.push(item)
+        finally:
+            self.lock.release()
+        BlockingPriorityQueue._notify(self.read_condition)
+
+    def peek(self, block: bool = True, timeout: int = None):
+        if len(self) == 0:
+            if block:  # Block and wait till not empty
+                BlockingPriorityQueue._wait(self.read_condition, timeout)
+        self.lock.acquire()
+        try:
+            item = self.priority_queue.peek()
+        finally:
+            self.lock.release()
+        return item
+
+    def pop(self, block: bool = True, timeout: int = None):
+        if len(self) == 0:
+            if block:  # Block and wait till not empty
+                BlockingPriorityQueue._wait(self.read_condition, timeout)
+        self.lock.acquire()
+        try:
+            item = self.priority_queue.pop()
+        finally:
+            self.lock.release()
+        BlockingPriorityQueue._notify(self.write_condition)
+        return item
+
+
 
 
